from nautobot.apps.ui import TemplateExtension

from .models import MeasurementLog
from .models import CableInventoryItem


class CableCommission(TemplateExtension):
    model = "dcim.cable"

    def buttons(self):
        cable = self.context["object"]
        log = None

        maybe_log = MeasurementLog.objects.filter(cable=cable)
        if maybe_log.exists():
            log = maybe_log.first().link

        cable_inventory_item = CableInventoryItem.objects.filter(cable=cable)

        return self.render(
            "nautobot_cable_utils/inc/buttons.html",
            {
                "cable": cable,
                "cable_planned": cable.status.name == "Planned",
                "cable_inventory": cable_inventory_item.exists(),
                "log": log,
            },
        )


class DeviceBulkConnect(TemplateExtension):
    model = "dcim.device"

    def buttons(self):
        device = self.context["object"]
        cable_available = None

        if device.get_cables().filter(status__name="Planned").count() > 0:
            cable_available = (
                device.get_cables().filter(status__name="Planned").first().pk
            )

        return self.render(
            "nautobot_cable_utils/inc/buttons_device.html",
            {
                "device": device,
                "cable_planned": cable_available is not None,
            },
        )


class InterfaceAutoRouteStart(TemplateExtension):
    model = "dcim.interface"

    def buttons(self):
        return self.render(
            "nautobot_cable_utils/inc/buttons_interface.html",
        )


class FrontPortAutoRouteStart(TemplateExtension):
    model = "dcim.frontport"

    def buttons(self):
        return self.render(
            "nautobot_cable_utils/inc/buttons_frontport.html",
        )


class CableInventoryItemPanel(TemplateExtension):
    model = "dcim.cable"

    def right_page(self):
        return self.render("nautobot_cable_utils/inc/cableinventoryitem_panel.html")


template_extensions = [
    CableCommission,
    DeviceBulkConnect,
    InterfaceAutoRouteStart,
    FrontPortAutoRouteStart,
    CableInventoryItemPanel,
]



from django import forms
from django.forms import Form, ChoiceField, CharField, IntegerField, BooleanField

from nautobot.circuits.models import Circuit, CircuitTermination, Provider
from nautobot.dcim.choices import CableTypeChoices, CableLengthUnitChoices
from nautobot.dcim.models import (
    Cable,
    ConsolePort,
    ConsoleServerPort,
    Device,
    FrontPort,
    Interface,
    PowerFeed,
    PowerOutlet,
    PowerPanel,
    PowerPort,
    Rack,
    RearPort,
    Manufacturer,
    Location,
)

from nautobot.extras.models import Status
from nautobot.apps.choices import ChoiceSet

from nautobot.apps.forms import (
    BootstrapMixin,
    ColorSelect,
    CSVModelForm,
    CSVModelChoiceField,
    DynamicModelChoiceField,
    StaticSelect2,
    add_blank_choice,
    parse_alphanumeric_range,
    DynamicModelMultipleChoiceField,
    NautobotFilterForm,
    NautobotBulkEditForm,
)
from nautobot.tenancy.models import Tenant
from .cable_router import CableRouter

from .models import CableInventoryItem, MeasurementLog, CablePlug
from django.core.exceptions import ObjectDoesNotExist, ValidationError
import re

NUMERIC_EXPANSION_PATTERN = r"\[((?:\d+[?:,-])+\d+)\]"


def expand_alphanumeric_pattern(string):
    """
    Expand an alphabetic pattern into a list of strings.
    """
    lead, pattern, remnant = re.split(NUMERIC_EXPANSION_PATTERN, string, maxsplit=1)
    parsed_range = parse_alphanumeric_range(pattern)

    maximum_length = max([len(str(x)) for x in parsed_range])

    for i in parsed_range:
        padded_i = str(i).zfill(maximum_length)
        if re.search(NUMERIC_EXPANSION_PATTERN, remnant):
            for string2 in expand_alphanumeric_pattern(remnant):
                yield f"{lead}{padded_i}{string2}"
        else:
            yield f"{lead}{padded_i}{remnant}"


class ExpandableCableNumberPatternField(forms.CharField):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.help_text:
            self.help_text = """
                Numeric ranges are supported for bulk creation. Mixed cases and types within a single range
                are not supported. Leading zeros will be added. Examples:
                <ul>
                    <li><code>e[0-3][0-13,17]</code></li>
                </ul>
                """

    def to_python(self, value):
        if not value:
            return ""
        if re.search(NUMERIC_EXPANSION_PATTERN, value):
            return list(expand_alphanumeric_pattern(value))
        return [value]


class ConnectCableToDeviceForm(BootstrapMixin, forms.ModelForm):
    """
    Base form for connecting a Cable to a Device component
    """

    termination_a_location = DynamicModelChoiceField(
        queryset=Location.objects.all(),
        label="Location",
        required=False,
    )
    termination_a_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label="Rack",
        required=False,
        display_field="display_name",
        null_option="None",
        query_params={"location": "$termination_a_location"},
    )
    termination_a_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label="Device",
        required=False,
        display_field="display_name",
        query_params={
            "location": "$termination_a_location",
            "rack": "$termination_a_rack",
        },
    )
    termination_b_location = DynamicModelChoiceField(
        queryset=Location.objects.all(), label="Location", required=False
    )
    termination_b_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label="Rack",
        required=False,
        display_field="display_name",
        null_option="None",
        query_params={"location": "$termination_b_location"},
    )
    termination_b_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label="Device",
        required=False,
        display_field="display_name",
        query_params={
            "location": "$termination_b_location",
            "rack": "$termination_b_rack",
        },
    )

    class Meta:
        model = Cable
        fields = [
            "termination_a_location",
            "termination_a_rack",
            "termination_a_device",
            "termination_a_id",
            "termination_b_location",
            "termination_b_rack",
            "termination_b_device",
            "termination_b_id",
        ]

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_a_id"], "pk", None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_b_id"], "pk", None)


class ConnectConsolePortToConsoleServerPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsolePort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=ConsoleServerPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectConsolePortToFrontPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsolePort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectConsolePortToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsolePort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectConsoleServerPortToFrontPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsoleServerPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectConsoleServerPortToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=ConsoleServerPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectPowerfeedToPowerPortForm(ConnectCableToDeviceForm):
    termination_a_location = DynamicModelChoiceField(
        queryset=Location.objects.all(),
        label="Location",
        required=False,
        display_field="cid",
    )
    termination_a_powerpanel = DynamicModelChoiceField(
        queryset=PowerPanel.objects.all(),
        label="Power Panel",
        required=False,
        query_params={
            "location": "$termination_a_location",
        },
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=PowerFeed.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"power_panel_id": "$termination_a_powerpanel"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=PowerPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )

    class Meta:
        model = Cable
        fields = [
            "termination_a_location",
            "termination_a_powerpanel",
            "termination_a_id",
            "termination_b_location",
            "termination_b_rack",
            "termination_b_device",
            "termination_b_id",
        ]


class ConnectPowerOutletToPowerPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=PowerOutlet.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=PowerPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectInterfaceForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={
            "device": "$termination_a_device",
            "kind": "physical",
        },
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={
            "device": "$termination_b_device",
            "kind": "physical",
        },
    )


class ConnectFrontPortToInterfaceForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={
            "device": "$termination_b_device",
            "kind": "physical",
        },
    )


class ConnectInterfaceToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={
            "device": "$termination_a_device",
            "kind": "physical",
        },
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectCircuitTerminationToInterfaceForm(ConnectCableToDeviceForm):
    termination_a_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(), label="Provider", required=False
    )
    termination_a_location = DynamicModelChoiceField(
        queryset=Location.objects.all(), label="Location", required=False
    )
    termination_a_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label="Circuit",
        display_field="cid",
        query_params={
            "provider_id": "$termination_a_provider",
            "location": "$termination_a_location",
        },
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label="Side",
        display_field="term_side",
        disabled_indicator="cable",
        query_params={"circuit_id": "$termination_a_circuit"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={
            "device": "$termination_b_device",
            "kind": "physical",
        },
    )

    class Meta:
        model = Cable
        fields = [
            "termination_a_provider",
            "termination_a_location",
            "termination_a_circuit",
            "termination_a_id",
            "termination_b_location",
            "termination_b_rack",
            "termination_b_device",
            "termination_b_id",
        ]


class ConnectCircuitTerminationForm(BootstrapMixin, forms.ModelForm):
    termination_a_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(), label="Provider", required=False
    )
    termination_a_location = DynamicModelChoiceField(
        queryset=Location.objects.all(), label="Location", required=False
    )
    termination_a_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label="Circuit",
        display_field="cid",
        query_params={
            "provider_id": "$termination_a_provider",
            "location": "$termination_a_location",
        },
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label="Side",
        display_field="term_side",
        disabled_indicator="cable",
        query_params={"circuit_id": "$termination_a_circuit"},
    )
    termination_b_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(), label="Provider", required=False
    )
    termination_b_location = DynamicModelChoiceField(
        queryset=Location.objects.all(), label="Location", required=False
    )
    termination_b_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label="Circuit",
        display_field="cid",
        query_params={
            "provider_id": "$termination_b_provider",
            "location": "$termination_b_location",
        },
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label="Side",
        display_field="term_side",
        disabled_indicator="cable",
        query_params={"circuit_id": "$termination_b_circuit"},
    )

    class Meta:
        model = Cable
        fields = [
            "termination_a_provider",
            "termination_a_location",
            "termination_a_circuit",
            "termination_a_id",
            "termination_b_provider",
            "termination_b_location",
            "termination_b_circuit",
            "termination_b_id",
            "type",
            "status",
            "label",
            "color",
            "length",
            "length_unit",
        ]

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_a_id"], "pk", None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_b_id"], "pk", None)


class ConnectFrontPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectFrontPortToRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectRearPortForm(ConnectCableToDeviceForm):
    termination_a_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_a_device"},
    )
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectCircuitTerminationToForm(BootstrapMixin, forms.ModelForm):
    termination_a_provider = DynamicModelChoiceField(
        queryset=Provider.objects.all(), label="Provider", required=False
    )
    termination_a_location = DynamicModelChoiceField(
        queryset=Location.objects.all(), label="Location", required=False
    )
    termination_a_circuit = DynamicModelChoiceField(
        queryset=Circuit.objects.all(),
        label="Circuit",
        display_field="cid",
        query_params={
            "provider_id": "$termination_a_provider",
            "location": "$termination_a_location",
        },
    )
    termination_a_id = DynamicModelChoiceField(
        queryset=CircuitTermination.objects.all(),
        label="Side",
        display_field="term_side",
        disabled_indicator="cable",
        query_params={"circuit_id": "$termination_a_circuit"},
    )
    termination_b_location = DynamicModelChoiceField(
        queryset=Location.objects.all(), label="Location", required=False
    )
    termination_b_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label="Rack",
        required=False,
        display_field="display_name",
        null_option="None",
        query_params={"location": "$termination_b_location"},
    )
    termination_b_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label="Device",
        required=False,
        display_field="display_name",
        query_params={
            "location": "$termination_b_location",
            "rack": "$termination_b_rack",
        },
    )

    class Meta:
        model = Cable
        fields = [
            "termination_a_provider",
            "termination_a_location",
            "termination_a_circuit",
            "termination_a_id",
            "termination_b_location",
            "termination_b_rack",
            "termination_b_device",
            "termination_b_id",
        ]

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_a_id"], "pk", None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_b_id"], "pk", None)


class ConnectCircuitTerminationToFrontPortForm(ConnectCircuitTerminationToForm):
    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class ConnectCircuitTerminationToRearPortForm(ConnectCircuitTerminationToForm):
    termination_b_id = DynamicModelChoiceField(
        queryset=RearPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={"device": "$termination_b_device"},
    )


class CableInventoryItemForm(BootstrapMixin, forms.ModelForm):
    name = forms.CharField(label="Cable Number")

    class Meta:
        model = CableInventoryItem
        fields = (
            "name",
            "label",
            "owner",
            "type",
            "plug",
            "supplier",
            "storage_rack",
            "procurement_ident",
            "color",
            "length",
            "length_unit",
        )


class CableInventoryItemBulkCreateForm(BootstrapMixin, forms.Form):
    pattern = ExpandableCableNumberPatternField(label="Cable Number Pattern")


class CableInventoryItemBulkForm(CableInventoryItemForm):

    def save(self, commit=True):
        if not self.instance.label:
            self.instance.label = self.instance.name

        return super().save(commit)


class CableInventoryItemBulkEditForm(NautobotBulkEditForm):
    pk = forms.ModelMultipleChoiceField(
        queryset=CableInventoryItem.objects.all(), widget=forms.MultipleHiddenInput
    )

    type = forms.ChoiceField(
        choices=add_blank_choice(CableTypeChoices),
        required=False,
        widget=StaticSelect2(),
    )
    color = forms.CharField(max_length=6, required=False, widget=ColorSelect())

    owner = DynamicModelChoiceField(
        queryset=Tenant.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    plug = DynamicModelChoiceField(
        queryset=CablePlug.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    supplier = DynamicModelChoiceField(
        queryset=Manufacturer.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    storage_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    procurement_ident = CharField(
        max_length=100, required=False, label="Procurement Identifier"
    )

    length = IntegerField(
        required=False,
        min_value=0,
    )
    length_unit = forms.ChoiceField(
        choices=add_blank_choice(CableLengthUnitChoices),
        required=False,
        widget=StaticSelect2(),
    )

    class Meta:
        nullable_fields = [
            "type",
            "owner",
            "color",
            "plug",
            "supplier",
            "storage_rack",
            "procurement_ident",
            "length",
            "length_unit",
        ]


class CableInventoryItemFilterForm(NautobotFilterForm):
    model = CableInventoryItem
    q = forms.CharField(required=False, label="Search")
    name = forms.CharField(required=False, label="Cable Number")
    type = forms.MultipleChoiceField(
        choices=add_blank_choice(CableTypeChoices),
        required=False,
        widget=StaticSelect2(),
    )
    color = forms.CharField(max_length=6, required=False, widget=ColorSelect())

    owner = DynamicModelMultipleChoiceField(
        queryset=Tenant.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    plug = DynamicModelMultipleChoiceField(
        queryset=CablePlug.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    supplier = DynamicModelMultipleChoiceField(
        queryset=Manufacturer.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    storage_rack = DynamicModelMultipleChoiceField(
        queryset=Rack.objects.all(),
        to_field_name="name",
        required=False,
        null_option="None",
    )
    procurement_ident = CharField(
        max_length=100, required=False, label="Procurement Identifier"
    )


class CommissionForm(BootstrapMixin, forms.Form):
    cable = forms.ModelChoiceField(
        queryset=Cable.objects.all(),
        required=False,
        disabled=True,
        widget=forms.HiddenInput(),
    )
    inventory_item = forms.CharField(
        required=False,
    )

    def clean(self):
        super().clean()
        try:
            if (
                CableInventoryItem.objects.get(
                    name=self.cleaned_data["inventory_item"]
                ).cable
                is not None
            ):
                raise forms.ValidationError(
                    {
                        "inventory_item": "the assigned cable inventory_item is already in use"
                    }
                )
        except ObjectDoesNotExist:
            raise forms.ValidationError(
                {"inventory_item": "the assigned cable inventory_item does not exist"}
            )

    def save(self):

        inventory_item = CableInventoryItem.objects.get(
            name=self.cleaned_data["inventory_item"]
        )
        cable = self.cleaned_data["cable"]

        for field in ["type", "label", "color", "length", "length_unit"]:
            value = getattr(inventory_item, field)
            if value:
                setattr(cable, field, value)

        if not inventory_item.label:
            cable.label = inventory_item.name

        connected_status = None
        statuses = Status.objects.get_for_model(Cable)

        for status in statuses:
            if status.name == "Connected":
                connected_status = status

        cable.status = connected_status
        cable.save()
        inventory_item.cable = cable
        inventory_item.save()

        return cable


class CableInventoryItemCSVForm(CSVModelForm):
    owner = CSVModelChoiceField(
        queryset=Tenant.objects.all(),
        required=False,
        to_field_name="name",
        help_text="Assigned Owner",
    )
    plug = CSVModelChoiceField(
        queryset=CablePlug.objects.all(),
        required=False,
        to_field_name="name",
        help_text="Plug",
    )
    supplier = CSVModelChoiceField(
        queryset=Manufacturer.objects.all(),
        required=False,
        to_field_name="name",
        help_text="Supplier",
    )

    class Meta:
        model = CableInventoryItem
        fields = CableInventoryItem.import_csv_headers


class MeasurementLogForm(BootstrapMixin, forms.ModelForm):
    cable = DynamicModelChoiceField(queryset=Cable.objects.all(), display_field="label")

    class Meta:
        model = MeasurementLog
        fields = ("link", "cable")


class CableRouterMediaChoices(ChoiceSet):
    FIBER_SM = "fiber_sm"
    FIBER_MM = "fiber_mm"
    COPPER = "copper"
    CHOICES = (
        (FIBER_SM, "Single-Mode Fiber"),
        (FIBER_MM, "Multi-Mode Fiber"),
        (COPPER, "Copper"),
    )


class CableRouterLinkForm(BootstrapMixin, Form):
    """
    Base form for connecting a Cable to a Device component
    """

    termination_a_type = None
    termination_b_type = None
    termination_a = None

    def __init__(self, *args, **kwargs):
        self.termination_a = kwargs.pop("termination_a")
        self.termination_a_type = kwargs.pop("termination_a_type")
        self.termination_b_type = kwargs.pop("termination_b_type")
        super().__init__(*args, **kwargs)

    termination_b_location = DynamicModelChoiceField(
        queryset=Location.objects.all(), label="Location", required=False
    )
    termination_b_rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        label="Rack",
        required=False,
        display_field="display_name",
        null_option="None",
        query_params={"location": "$termination_b_location"},
    )
    termination_b_device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        label="Device",
        required=False,
        display_field="display_name",
        query_params={
            "location": "$termination_b_location",
            "rack": "$termination_b_rack",
        },
    )

    media_type = ChoiceField(choices=CableRouterMediaChoices, label="Media Type")

    allow_next_rack_hops = BooleanField(
        label="Allow Next Rack Patches", initial=True, required=False
    )

    class Meta:
        fields = [
            "termination_a_location",
            "termination_a_rack",
            "termination_a_device",
            "termination_a_id",
            "termination_b_location",
            "termination_b_rack",
            "termination_b_device",
            "termination_b_id",
        ]

    def clean(self):
        cleaned_data = super().clean()

        termination_b_model = self.termination_b_type.model_class()
        termination_b = termination_b_model.objects.get(
            id=cleaned_data["termination_b_id"]
        )
        media_type = self.cleaned_data["media_type"]

        cleaned_data["cable_router"] = CableRouter(
            self.termination_a,
            self.termination_a_type,
            termination_b,
            self.termination_b_type,
            media_type,
            enable_next_rack_hops=self.cleaned_data["allow_next_rack_hops"],
        )
        path = cleaned_data["cable_router"].get_path()
        if path is None:
            raise ValidationError(
                f"{self.termination_a.device} and {termination_b.device} are not connected"
            )

        cleaned_data["path"] = path

    def clean_termination_a_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_a_id"], "pk", None)

    def clean_termination_b_id(self):
        # Return the PK rather than the object
        return getattr(self.cleaned_data["termination_b_id"], "pk", None)


class CableRouterLinkToInterfaceForm(CableRouterLinkForm):
    termination_b_model = Interface

    termination_b_id = DynamicModelChoiceField(
        queryset=Interface.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={
            "device": "$termination_b_device",
            "kind": "physical",
        },
    )

    class Meta(CableRouterLinkForm.Meta):
        pass


class CableRouterLinkToFrontPortForm(CableRouterLinkForm):
    termination_b_model = FrontPort

    termination_b_id = DynamicModelChoiceField(
        queryset=FrontPort.objects.all(),
        label="Name",
        disabled_indicator="cable",
        query_params={
            "device": "$termination_b_device",
        },
    )

    class Meta(CableRouterLinkForm.Meta):
        pass


class CableRouterVerifyForm(BootstrapMixin, Form):
    class Meta:
        fields = []



from nautobot.apps.ui import (
    ButtonColorChoices,
    NavMenuTab,
    NavMenuGroup,
    NavMenuItem,
    NavMenuButton,
)


menu_items = (
    NavMenuTab(
        name="Plugins",
        groups=[
            NavMenuGroup(
                name="Cable Utilities",
                items=[
                    NavMenuItem(
                        link="plugins:nautobot_cable_utils:cableinventoryitem_list",
                        name="Cable Inventory",
                        permissions=["dcim.view_cables"],
                        buttons=[
                            NavMenuButton(
                                link="plugins:nautobot_cable_utils:cableinventoryitem_add",
                                title="Add Cable Inventory Item",
                                icon_class="mdi mdi-plus-thick",
                                button_class=ButtonColorChoices.GREEN,
                            ),
                            NavMenuButton(
                                link="plugins:nautobot_cable_utils:cableinventoryitem_import",
                                title="Import Cable Inventory Items",
                                icon_class="mdi mdi-database-import-outline",
                                button_class=ButtonColorChoices.BLUE,
                            ),
                        ],
                    ),
                    NavMenuItem(
                        link="plugins:nautobot_cable_utils:measurement_log_list",
                        name="Measurement Logs",
                        permissions=["dcim.view_cables"],
                        buttons=[
                            NavMenuButton(
                                link="plugins:nautobot_cable_utils:measurement_log_add",
                                title="Add Measurement Log",
                                icon_class="mdi mdi-plus-thick",
                                button_class=ButtonColorChoices.GREEN,
                            ),
                        ],
                    ),
                ],
            ),
        ],
    ),
)



from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views.generic import View, FormView

from nautobot.apps.views import (
    BulkCreateView,
    BulkDeleteView,
    BulkEditView,
    BulkImportView,
    ObjectDeleteView,
    GetReturnURLMixin,
    ObjectEditView,
    ObjectListView,
    ObjectView,
)
from nautobot.dcim.models import Cable, Device, Interface, FrontPort
from nautobot.apps.forms import ConfirmationForm
from nautobot.extras.models.statuses import Status

from . import filters, forms, tables
from .cable_router import CablePath
from .forms import (
    CableRouterVerifyForm,
    CableRouterLinkToInterfaceForm,
    CableRouterLinkToFrontPortForm,
)
from .models import CableInventoryItem, MeasurementLog
from .tables import NeededCableTable


class ReconnectView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = "dcim.add_cable"
    template_name = "nautobot_cable_utils/cable_connect.html"

    def dispatch(self, request, *args, pk=None, **kwargs):
        self.obj = Cable.objects.get(pk=pk)

        idx = (self.obj.termination_a_type.model, self.obj.termination_b_type.model)
        idx_sorted = tuple(sorted(idx))
        if idx != idx_sorted:
            termination_a = self.obj.termination_a
            self.obj.termination_a = self.obj.termination_b
            self.obj.termination_b = termination_a

        self.form_class = {
            (
                "circuittermination",
                "circuittermination",
            ): forms.ConnectCircuitTerminationForm,
            (
                "circuittermination",
                "rearport",
            ): forms.ConnectCircuitTerminationToRearPortForm,
            (
                "consoleport",
                "consoleserverport",
            ): forms.ConnectConsolePortToConsoleServerPortForm,
            ("consoleport", "frontport"): forms.ConnectConsolePortToFrontPortForm,
            ("consoleport", "rearport"): forms.ConnectConsolePortToRearPortForm,
            (
                "consoleserverport",
                "frontport",
            ): forms.ConnectConsoleServerPortToFrontPortForm,
            (
                "consoleserverport",
                "rearport",
            ): forms.ConnectConsoleServerPortToRearPortForm,
            ("powerfeed", "powerport"): forms.ConnectPowerfeedToPowerPortForm,
            ("poweroutlet", "powerport"): forms.ConnectPowerOutletToPowerPortForm,
            (
                "circuittermination",
                "interface",
            ): forms.ConnectCircuitTerminationToInterfaceForm,
            ("frontport", "interface"): forms.ConnectFrontPortToInterfaceForm,
            ("interface", "rearport"): forms.ConnectInterfaceToRearPortForm,
            ("interface", "interface"): forms.ConnectInterfaceForm,
            ("frontport", "frontport"): forms.ConnectFrontPortForm,
            ("frontport", "rearport"): forms.ConnectFrontPortToRearPortForm,
            (
                "circuittermination",
                "frontport",
            ): forms.ConnectCircuitTerminationToFrontPortForm,
            ("rearport", "rearport"): forms.ConnectRearPortForm,
        }[idx_sorted]

        return super().dispatch(request, *args, **kwargs)

    def prefill_form(self, initial_data, termination):
        o = getattr(self.obj, termination)
        if o and hasattr(o, "device"):
            device = o.device
            initial_data["{}_device".format(termination)] = device
            if device.location:
                initial_data["{}_location".format(termination)] = device.location
            if device.rack:
                initial_data["{}_rack".format(termination)] = device.rack

    def get(self, request, *args, **kwargs):
        # Parse initial data manually to avoid setting field values as lists
        initial_data = {k: request.GET[k] for k in request.GET}

        self.prefill_form(initial_data, "termination_a")
        self.prefill_form(initial_data, "termination_b")

        form = self.form_class(instance=self.obj, initial=initial_data)

        return render(
            request,
            self.template_name,
            {
                "obj": self.obj,
                "obj_type": Cable._meta.verbose_name,
                "form": form,
                "return_url": self.get_return_url(request, self.obj),
            },
        )

    def post(self, request, *args, **kwargs):
        assigned_cable_inventory_item = None

        with transaction.atomic():
            try:
                if (
                    hasattr(self.obj.termination_a, "_path")
                    and self.obj.termination_a._path_id
                ):
                    self.obj.termination_a._path = None
            except CablePath.DoesNotExist:
                pass
            try:
                if (
                    hasattr(self.obj.termination_b, "_path")
                    and self.obj.termination_b._path_id
                ):
                    self.obj.termination_b._path = None
            except CablePath.DoesNotExist:
                pass

            assigned_cable_inventory_item = CableInventoryItem.objects.filter(
                cable=self.obj
            ).first()

            self.obj.delete()
            self.obj.pk = None
            self.obj._state.adding = True

            self.obj.termination_a.cable = None
            self.obj.termination_b.cable = None

            form = self.form_class(request.POST, request.FILES, instance=self.obj)
            if form.is_valid():
                obj = form.save()

                if assigned_cable_inventory_item is not None:
                    # restore cable template if previously assigned
                    assigned_cable_inventory_item.cable = obj
                    assigned_cable_inventory_item.save()

                redirect_url = self.get_return_url(request, obj)
                return redirect(redirect_url)

            return render(
                request,
                self.template_name,
                {
                    "obj": self.obj,
                    "obj_type": Cable._meta.verbose_name,
                    "form": form,
                    "return_url": self.get_return_url(request, self.obj),
                },
            )


class CommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = "dcim.edit_cable"
    template_name = "nautobot_cable_utils/commission_cable.html"

    def get(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = forms.CommissionForm(initial={"cable": cable})

        return render(
            request,
            self.template_name,
            {
                "form": form,
                "cable": cable,
                "return_url": self.get_return_url(request, cable),
            },
        )

    def post(self, request, pk=None):
        with transaction.atomic():
            cable = get_object_or_404(Cable, pk=pk)
            form = forms.CommissionForm(request.POST, initial={"cable": cable})

            if form.is_valid():
                form.save()

                return redirect(self.get_return_url(request, cable))

            return render(
                request,
                self.template_name,
                {
                    "form": form,
                    "cable": cable,
                    "return_url": self.get_return_url(request, cable),
                },
            )


class UndoCommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = "dcim.edit_cable"
    queryset = Cable.objects.all()
    template_name = "nautobot_cable_utils/undo_commission.html"

    def get(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = ConfirmationForm(initial=request.GET)

        return render(
            request,
            self.template_name,
            {
                "obj": cable,
                "form": form,
                "obj_type": self.queryset.model._meta.verbose_name,
                "return_url": self.get_return_url(request, cable),
            },
        )

    def post(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = ConfirmationForm(request.POST)

        if form.is_valid():
            with transaction.atomic():
                cable_new = Cable()
                cable_new.termination_a = cable.termination_a
                cable_new.termination_b = cable.termination_b
                cable_new.status = Status.objects.get(name="Planned")

                cable.delete()
                cable_new.save()

        return redirect(self.get_return_url(request, cable_new))


class DeviceCommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = "dcim.edit_cable"
    template_name = "nautobot_cable_utils/bulk_commission.html"

    def get(self, request, pk=None):
        device = get_object_or_404(Device, pk=pk)

        # set key, if not present
        if "current_device" not in request.session:
            request.session["current_device"] = None

        if request.session["current_device"] != str(pk):
            cable_list = list()
            for cable in device.get_cables().filter(status__name="Planned"):
                cable_list.append(str(cable.pk))

            request.session["current_device"] = str(pk)
            request.session["cable_list"] = cable_list
            request.session.modified = True

        cable = get_object_or_404(Cable, pk=request.session["cable_list"][0])
        form = forms.CommissionForm(initial={"cable": cable})

        return render(
            request,
            self.template_name,
            {
                "form": form,
                "count": len(request.session["cable_list"]),
                "object": cable,
                "verbose_name": Cable._meta.verbose_name,
                "verbose_name_plural": Cable._meta.verbose_name_plural,
                "return_url": self.get_return_url(request, device),
                "active_tab": "main",
            },
        )

    def post(self, request, pk=None):
        with transaction.atomic():
            device = get_object_or_404(Device, pk=pk)
            cable = get_object_or_404(Cable, pk=request.session["cable_list"][0])
            form = forms.CommissionForm(request.POST, initial={"cable": cable})

            if "_create" in self.request.POST:
                if form.is_valid():
                    request.session["cable_list"].pop(0)
                    request.session.modified = True
                    form.save()
            elif "_skip" in self.request.POST:
                request.session["cable_list"].pop(0)
                request.session.modified = True

            if (
                len(request.session["cable_list"]) > 0
                and "_cancel" not in self.request.POST
            ):
                if form.is_valid() or "_skip" in self.request.POST:
                    return redirect(request.path)
                else:
                    # re-render page to display ValidationErrors
                    return render(
                        request,
                        self.template_name,
                        {
                            "form": form,
                            "count": len(request.session["cable_list"]),
                            "object": get_object_or_404(
                                Cable, pk=request.session["cable_list"][0]
                            ),
                            "verbose_name": Cable._meta.verbose_name,
                            "verbose_name_plural": Cable._meta.verbose_name_plural,
                            "return_url": self.get_return_url(request, device),
                            "active_tab": "main",
                        },
                    )

            if "current_device" in self.request.session.keys():
                del request.session["current_device"]
            if "cable_list" in self.request.session.keys():
                del request.session["cable_list"]

        return redirect(self.get_return_url(request, device))


class CableInventoryItemListView(ObjectListView):
    queryset = CableInventoryItem.objects.all()
    filterset = filters.CableInventoryItemFilterSet
    filterset_form = forms.CableInventoryItemFilterForm
    table = tables.CableInventoryItemTable


class CableInventoryItemView(ObjectView):
    queryset = CableInventoryItem.objects.all()


class CableInventoryItemCreateView(ObjectEditView):
    queryset = CableInventoryItem.objects.all()
    model_form = forms.CableInventoryItemForm
    default_return_url = "plugins:nautobot_cable_utils:cableinventoryitem_list"
    template_name = "nautobot_cable_utils/cableinventoryitem_edit.html"


class CableInventoryItemBulkCreateView(BulkCreateView):
    queryset = CableInventoryItem.objects.all()
    form = forms.CableInventoryItemBulkCreateForm
    model_form = forms.CableInventoryItemBulkForm
    pattern_target = "name"
    template_name = "nautobot_cable_utils/cableinventoryitem_bulk_add.html"


class CableInventoryItemBulkEditView(BulkEditView):
    queryset = CableInventoryItem.objects.all()
    form = forms.CableInventoryItemBulkEditForm
    model_form = forms.CableInventoryItemBulkForm
    table = tables.CableInventoryItemTable


class CableInventoryItemEditView(CableInventoryItemCreateView):
    pass


class CableInventoryItemBulkImportView(BulkImportView):
    queryset = CableInventoryItem.objects.all()
    model_form = forms.CableInventoryItemCSVForm
    table = tables.CableInventoryItemTable


class CableInventoryItemBulkDeleteView(BulkDeleteView):
    queryset = CableInventoryItem.objects.all()
    filterset = filters.CableInventoryItemFilterSet
    table = tables.CableInventoryItemTable


class CableInventoryItemDeleteView(ObjectDeleteView):
    queryset = CableInventoryItem.objects.all()


class MeasurementLogListView(ObjectListView):
    queryset = MeasurementLog.objects.all()
    table = tables.MeasurementLogTable
    action_buttons = tuple()
    template_name = "nautobot_cable_utils/measurement_log_list.html"


class MeasurementLogCreateView(ObjectEditView):
    queryset = MeasurementLog.objects.all()
    model_form = forms.MeasurementLogForm
    default_return_url = "plugins:nautobot_cable_utils:measurement_log_list"


class MeasurementLogEditView(CableInventoryItemCreateView):
    pass


class MeasurementLogBulkDeleteView(BulkDeleteView):
    queryset = MeasurementLog.objects.all()
    table = tables.MeasurementLogTable


class CableRouterAddView(PermissionRequiredMixin, FormView):
    permission_required = "dcim.add_cable"
    template_name = "nautobot_cable_utils/cable_router_add.html"

    termination_types = {
        "interface": ContentType.objects.get_for_model(Interface),
        "front-port": ContentType.objects.get_for_model(FrontPort),
    }

    termination_forms = {
        "interface": CableRouterLinkToInterfaceForm,
        "front-port": CableRouterLinkToFrontPortForm,
    }

    termination_names = {
        "interface": "Interface",
        "front-port": "Front Port",
    }

    termination_a_type = None
    termination_b_type = None
    termination_a_type_name = None
    termination_b_type_name = None
    termination_a = None
    form_class = None

    def dispatch(self, request, *args, **kwargs):
        self.termination_a_type = self.termination_types[
            kwargs.get("termination_a_type")
        ]
        self.termination_b_type = self.termination_types[
            kwargs.get("termination_b_type")
        ]
        self.termination_a_type_name = self.termination_names[
            kwargs.get("termination_a_type")
        ]
        self.termination_b_type_name = self.termination_names[
            kwargs.get("termination_b_type")
        ]
        self.termination_a = self.termination_a_type.model_class().objects.get(
            id=kwargs.get("termination_a_id")
        )
        self.form_class = self.termination_forms[kwargs.get("termination_b_type")]

        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        return {
            **kwargs,
            "termination_a": self.termination_a,
            "termination_a_type": self.termination_a_type,
            "termination_b_type": self.termination_b_type,
        }

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data()

        return {
            **ctx,
            "termination_a_type": self.termination_a_type,
            "termination_b_type": self.termination_b_type,
            "termination_a_type_name": self.termination_a_type_name,
            "termination_b_type_name": self.termination_b_type_name,
            "termination_a": self.termination_a,
        }

    def form_valid(self, form):
        query_params1 = map(
            lambda x: f"cable={x['cable_id']}_{'r' if x['is_reversed'] else 'n'}",
            form.cleaned_data["path"],
        )
        query_params2 = [
            f"media_type={form.cleaned_data['media_type']}",
            f"termination_a_type={self.termination_a_type.id}",
            f"termination_a_id={self.termination_a.id}",
            f"termination_b_type={self.termination_b_type.id}",
            f"termination_b_id={form.cleaned_data['termination_b_id']}",
        ]
        query_params = "&".join([*query_params1, *query_params2])
        redirect_url = (
            reverse("plugins:nautobot_cable_utils:cable_router_verify")
            + "?"
            + query_params
        )
        return HttpResponseRedirect(redirect_url)


class CableRouterVerifyView(PermissionRequiredMixin, FormView):
    permission_required = "dcim.add_cable"
    form_class = CableRouterVerifyForm
    template_name = "nautobot_cable_utils/cable_router_verify.html"

    def get_cable_path(self):
        ct_termination_a = ContentType.objects.get_for_id(
            self.request.GET.get("termination_a_type")
        )
        ct_termination_b = ContentType.objects.get_for_id(
            self.request.GET.get("termination_b_type")
        )

        termination_a = ct_termination_a.model_class().objects.get(
            pk=self.request.GET.get("termination_a_id")
        )
        termination_b = ct_termination_b.model_class().objects.get(
            pk=self.request.GET.get("termination_b_id")
        )

        cable_list_raw = self.request.GET.getlist("cable")
        cable_path = CablePath(
            termination_a,
            ct_termination_a,
            termination_b,
            ct_termination_b,
            cable_list_raw,
        )

        return cable_path

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        cable_path = self.get_cable_path()

        needed_cable_table = NeededCableTable(cable_path.needed_cables)

        return {
            **ctx,
            "termination_a": cable_path.termination_a,
            "termination_b": cable_path.termination_b,
            "path": cable_path.get_renderable_path(),
            "needed_cable_table": needed_cable_table,
        }

    def form_valid(self, form):
        cable_path = self.get_cable_path()
        needed_cables = cable_path.needed_cables
        planned_status = Status.objects.get(name="Planned")

        for needed_cable in needed_cables:
            Cable.objects.create(
                termination_a=needed_cable["termination_a"],
                termination_b=needed_cable["termination_b"],
                status=planned_status,
            )

        messages.success(self.request, f"{len(needed_cables)} Cables were created.")

        return HttpResponseRedirect(cable_path.termination_a.get_absolute_url())
